@@ -1,12 +1,12 @@
 ;;; ellama.el --- Tool for interacting with LLMs -*- lexical-binding: t -*-
 
-;; Copyright (C) 2023  Free Software Foundation, Inc.
+;; Copyright (C) 2023, 2024  Free Software Foundation, Inc.
 
 ;; Author: Sergey Kostyaev <sskostyaev@gmail.com>
 ;; URL: http://github.com/s-kostyaev/ellama
 ;; Keywords: help local tools
-;; Package-Requires: ((emacs "28.1") (llm "0.6.0") (spinner "1.7.4") (dash "2.19.1"))
-;; Version: 0.7.7
+;; Package-Requires: ((emacs "28.1") (llm "0.6.0") (spinner "1.7.4"))
+;; Version: 0.8.13
 ;; SPDX-License-Identifier: GPL-3.0-or-later
 ;; Created: 8th Oct 2023
 
@@ -38,7 +38,7 @@
 (require 'json)
 (require 'llm)
 (require 'spinner)
-(require 'dash)
+(require 'info)
 (eval-when-compile (require 'rx))
 
 (defgroup ellama nil
@@ -55,10 +55,10 @@
   :group 'ellama
   :type 'string)
 
-(defcustom ellama-nick-prefix "**"
-  "User and assistant nick prefix in logs."
+(defcustom ellama-nick-prefix-depth 2
+  "Prefix depth."
   :group 'ellama
-  :type 'string)
+  :type 'integer)
 
 (defcustom ellama-language "English"
   "Language for ellama translation."
@@ -75,6 +75,16 @@
   :group 'ellama
   :type '(sexp :validate 'cl-struct-p))
 
+(defcustom ellama-chat-translation-enabled nil
+  "Enable chat translations."
+  :group 'ellama
+  :type 'boolean)
+
+(defcustom ellama-translation-provider nil
+  "LLM provider for chat translation."
+  :group 'ellama
+  :type '(sexp :validate 'cl-struct-p))
+
 (defcustom ellama-providers nil
   "LLM provider list for fast switching."
   :group 'ellama
@@ -88,6 +98,55 @@
 		      `(const ,(car type)))
 		    spinner-types)))
 
+(defcustom ellama-command-map
+  (let ((map (make-sparse-keymap)))
+    ;; code
+    (define-key map (kbd "c c") 'ellama-code-complete)
+    (define-key map (kbd "c a") 'ellama-code-add)
+    (define-key map (kbd "c e") 'ellama-code-edit)
+    (define-key map (kbd "c i") 'ellama-code-improve)
+    (define-key map (kbd "c r") 'ellama-code-review)
+    ;; summarize
+    (define-key map (kbd "s s") 'ellama-summarize)
+    (define-key map (kbd "s w") 'ellama-summarize-webpage)
+    ;; session
+    (define-key map (kbd "s l") 'ellama-load-session)
+    (define-key map (kbd "s r") 'ellama-session-rename)
+    (define-key map (kbd "s d") 'ellama-session-remove)
+    (define-key map (kbd "s a") 'ellama-session-switch)
+    ;; improve
+    (define-key map (kbd "i w") 'ellama-improve-wording)
+    (define-key map (kbd "i g") 'ellama-improve-grammar)
+    (define-key map (kbd "i c") 'ellama-improve-conciseness)
+    ;; make
+    (define-key map (kbd "m l") 'ellama-make-list)
+    (define-key map (kbd "m t") 'ellama-make-table)
+    (define-key map (kbd "m f") 'ellama-make-format)
+    ;; ask
+    (define-key map (kbd "a a") 'ellama-ask-about)
+    (define-key map (kbd "a i") 'ellama-chat)
+    (define-key map (kbd "a l") 'ellama-ask-line)
+    (define-key map (kbd "a s") 'ellama-ask-selection)
+    ;; text
+    (define-key map (kbd "t t") 'ellama-translate)
+    (define-key map (kbd "t b") 'ellama-translate-buffer)
+    (define-key map (kbd "t c") 'ellama-complete)
+    (define-key map (kbd "t e") 'ellama-chat-translation-enable)
+    (define-key map (kbd "t d") 'ellama-chat-translation-disable)
+    ;; define
+    (define-key map (kbd "d w") 'ellama-define-word)
+    ;; context
+    (define-key map (kbd "x b") 'ellama-context-add-buffer)
+    (define-key map (kbd "x f") 'ellama-context-add-file)
+    (define-key map (kbd "x s") 'ellama-context-add-selection)
+    (define-key map (kbd "x i") 'ellama-context-add-info-node)
+    ;; provider
+    (define-key map (kbd "p s") 'ellama-provider-select)
+    map)
+  "Keymap for ellama commands."
+  :group 'ellama
+  :type 'keymap)
+
 (defun ellama-setup-keymap ()
   "Set up the Ellama keymap and bindings."
   (interactive)
@@ -95,42 +154,10 @@
     (defvar ellama-keymap (make-sparse-keymap)
       "Keymap for Ellama Commands")
 
-    (define-key global-map (kbd ellama-keymap-prefix) ellama-keymap)
+    (when ellama-keymap-prefix
+      (define-key global-map (kbd ellama-keymap-prefix) ellama-command-map))))
 
-    (let ((key-commands
-	   '(;; code
-	     ("c c" ellama-code-complete "Code complete")
-	     ("c a" ellama-code-add "Code add")
-	     ("c e" ellama-code-edit "Code edit")
-	     ("c i" ellama-code-improve "Code improve")
-	     ("c r" ellama-code-review "Code review")
-	     ;; summarize
-	     ("s s" ellama-summarize "Summarize")
-	     ("s w" ellama-summarize-webpage "Summarize webpage")
-	     ;; improve
-	     ("i w" ellama-improve-wording "Improve wording")
-	     ("i g" ellama-improve-grammar "Improve grammar and spelling")
-	     ("i c" ellama-improve-conciseness "Improve conciseness")
-	     ;; make
-	     ("m l" ellama-make-list "Make list")
-	     ("m t" ellama-make-table "Make table")
-	     ("m f" ellama-make-format "Make format")
-	     ;; ask
-	     ("a a" ellama-ask-about "Ask about")
-	     ("a i" ellama-chat "Ask interactively")
-	     ("a l" ellama-ask-line "Ask current line")
-	     ("a s" ellama-ask-selection "Ask selection")
-	     ;; text
-	     ("t t" ellama-translate "Text translate")
-	     ("t c" ellama-complete "Text complete")
-	     ;; define
-	     ("d w" ellama-define-word "Define word")
-	     ;; provider
-	     ("p s" ellama-provider-select "Provider select"))))
-      (dolist (key-command key-commands)
-	(define-key ellama-keymap (kbd (car key-command)) (cadr key-command))))))
-
-(defcustom ellama-keymap-prefix "C-c e"
+(defcustom ellama-keymap-prefix nil
   "Key sequence for Ellama Commands."
   :type 'string
   :set (lambda (symbol value)
@@ -163,20 +190,22 @@
   :group 'ellama
   :type 'integer)
 
-(defcustom ellama-ask-about-prompt-template "Text:\n%s\nRegarding this text, %s"
-  "Prompt template for `ellama-ask-about'."
-  :group 'ellama
-  :type 'string)
+(defcustom ellama-naming-scheme 'ellama-generate-name-by-words
+  "How to name sessions.
+If you choose custom function, that function should accept PROVIDER, ACTION
+and PROMPT arguments.
 
-(defcustom ellama-translate-word-prompt-template "Translate %s to %s"
-  "Promp template for `ellama-translate' with single word."
-  :group 'ellama
-  :type 'string)
+PROVIDER is an llm provider.
 
-(defcustom ellama-translate-region-prompt-template "Translate the following text to %s:\n%s"
-  "Promp template for `ellama-translate' with active region."
+ACTION is a symbol, current command.
+
+PROMPT is a prompt string."
   :group 'ellama
-  :type 'string)
+  :type `(choice
+          (const :tag "By first N words of prompt" ellama-generate-name-by-words)
+          (const :tag "By current time" ellama-generate-name-by-time)
+	  (const :tag "By generating name with LLM based on prompt." ellama-generate-name-by-llm)
+          (function :tag "By custom function")))
 
 (defcustom ellama-define-word-prompt-template "Define %s"
   "Prompt template for `ellama-define-word'."
@@ -248,6 +277,26 @@
   :group 'ellama
   :type 'string)
 
+(defcustom ellama-get-name-template "I will get you user query, you should return short topic only, what this conversation about. NEVER respond to query itself. Topic must be short and concise.
+For example:
+Query: Why is sky blue?
+Topic: Blue sky
+
+Query: %s
+Topic:"
+  "Prompt template for `ellama-get-name'."
+  :group 'ellama
+  :type 'string)
+
+(defcustom ellama-translation-template "Translate this text to %s.
+Original text:
+%s
+Translation to %s:
+"
+  "Translation template."
+  :group 'ellama
+  :type 'string)
+
 (defcustom ellama-chat-done-callback nil
   "Callback that will be called on ellama chat response generation done.
 It should be a function with single argument generated text string."
@@ -307,33 +356,23 @@ Too low value can break generated code by splitting long comment lines."
   ;; Trim left first as `string-trim' trims from the right and ends up deleting all the code.
   (string-trim-right (string-trim-left text ellama--code-prefix) ellama--code-suffix))
 
-(defun ellama--fill-string (s)
-  "Fill string S."
-  (with-temp-buffer
-    (emacs-lisp-mode)
-    (insert s)
-    (fill-region (point-min) (point-max))
-    (buffer-substring (point-min) (point-max))))
-
 (defun ellama--fill-long-lines (text)
   "Fill long lines only in TEXT."
-  (--> text
-       (split-string it "\n")
-       (-map (lambda (el)
-	       (if (> (length el)
-		      ellama-long-lines-length)
-		   (ellama--fill-string el)
-		 el)) it)
-       (string-join it "\n")))
+  (with-temp-buffer
+    (insert (propertize text 'hard t))
+    (let ((fill-column ellama-long-lines-length)
+	  (use-hard-newlines t))
+      (fill-region (point-min) (point-max) nil t t))
+    (buffer-substring-no-properties (point-min) (point-max))))
 
 (defun ellama--translate-markdown-to-org-filter (text)
   "Filter to translate code blocks from markdown syntax to org syntax in TEXT.
 This filter contains only subset of markdown syntax to be good enough."
-  (->> text
+  (thread-last text
        ;; code blocks
-       (replace-regexp-in-string "^```\\(.+\\)$" "#+BEGIN_SRC \\1")
+       (replace-regexp-in-string "^[[:space:]]*```\\(.+\\)$" "#+BEGIN_SRC \\1")
        (replace-regexp-in-string "^<!-- language: \\(.+\\) -->\n```" "#+BEGIN_SRC \\1")
-       (replace-regexp-in-string "^```$" "#+END_SRC")
+       (replace-regexp-in-string "^[[:space:]]*```$" "#+END_SRC")
        ;; lists
        (replace-regexp-in-string "^\\* " "+ ")
        ;; bold
@@ -341,7 +380,7 @@ This filter contains only subset of markdown syntax to be good enough."
        (replace-regexp-in-string "__\\(.+?\\)__" "*\\1*")
        (replace-regexp-in-string "<b>\\(.+?\\)</b>" "*\\1*")
        ;; italic
-       (replace-regexp-in-string "_\\(.+?\\)_" "/\\1/")
+       ;; (replace-regexp-in-string "_\\(.+?\\)_" "/\\1/") ;; most of the time it breaks code blocks, so disable it
        (replace-regexp-in-string "<i>\\(.+?\\)</i>" "/\\1/")
        ;; inline code
        (replace-regexp-in-string "`\\(.+?\\)`" "~\\1~")
@@ -375,11 +414,6 @@ This filter contains only subset of markdown syntax to be good enough."
 	   ;; If ellama-enable-keymap is nil, remove the key bindings
 	   (define-key global-map (kbd ellama-keymap-prefix) nil))))
 
-(defcustom ellama-session-file-extension "org"
-  "File extension for saving ellama session."
-  :type 'string
-  :group 'ellama)
-
 (defcustom ellama-sessions-directory (file-truename
 				      (file-name-concat
 				       user-emacs-directory
@@ -388,6 +422,11 @@ This filter contains only subset of markdown syntax to be good enough."
   :type 'string
   :group 'ellama)
 
+(defcustom ellama-naming-provider nil
+  "LLM provider for generating names."
+  :group 'ellama
+  :type '(sexp :validate 'cl-struct-p))
+
 (defvar-local ellama--current-session nil)
 
 (defvar ellama--current-session-id nil)
@@ -403,15 +442,17 @@ PROVIDER is an llm provider of session.
 
 FILE is a path to file contains string representation of this session, string.
 
-PROMPT is a variable contains last prompt in this session."
-  id provider file prompt)
+PROMPT is a variable contains last prompt in this session.
+
+CONTEXT contains context for next request."
+  id provider file prompt context)
 
 (defun ellama-get-session-buffer (id)
   "Return ellama session buffer by provided ID."
   (gethash id ellama--active-sessions))
 
-(defun ellama-generate-name (provider action prompt)
-  "Generate name for ellama ACTION by PROVIDER according to PROMPT."
+(defun ellama-generate-name-by-words (provider action prompt)
+  "Generate name for ACTION by PROVIDER by getting first N words from PROMPT."
   (let* ((cleaned-prompt (replace-regexp-in-string "/" "_" prompt))
          (prompt-words (split-string cleaned-prompt)))
     (string-join
@@ -424,6 +465,54 @@ PROMPT is a variable contains last prompt in this session."
 	    (format "(%s)" (llm-name provider))))
      " ")))
 
+(defun ellama-get-name (prompt)
+  "Generate session name by LLM based on PROMPT."
+  (let ((provider (or ellama-naming-provider ellama-provider)))
+    (string-trim-right
+     (string-trim
+      (seq-first
+       (split-string
+	(llm-chat provider (llm-make-simple-chat-prompt
+			    (format ellama-get-name-template prompt)))
+	"\n")))
+     "\\.")))
+
+(defun ellama-generate-name-by-llm (provider _action prompt)
+  "Generate name for ellama ACTION by PROVIDER and PROMPT by LLM."
+  (format "%s (%s)"
+	  (ellama-get-name prompt)
+	  (llm-name provider)))
+
+(defun ellama-get-current-time ()
+  "Return string representation of current time."
+  (replace-regexp-in-string
+   "\\([0-9]\\{2\\}\\)\\([0-9]\\{2\\}\\)\\'" "\\1:\\2"
+   (format-time-string "%FT%T%z" (current-time))))
+
+(defun ellama-generate-name-by-time (_provider _action _prompt)
+  "Generate name for ellama session by current time."
+  (ellama-get-current-time))
+
+(defun ellama-generate-name (provider action prompt)
+  "Generate name for ellama ACTION by PROVIDER according to PROMPT."
+  (replace-regexp-in-string "/" "_" (funcall ellama-naming-scheme provider action prompt)))
+
+(defvar ellama--new-session-context nil)
+
+(defun ellama-get-nick-prefix-for-mode ()
+  "Return preferred header prefix char based om the current mode.
+Defaults to #, but supports `org-mode'.  Depends on `ellama-major-mode'."
+  (let* ((prefix-char
+          (cond ((provided-mode-derived-p ellama-major-mode 'org-mode) ?*)
+                (t ?#))))
+    (make-string ellama-nick-prefix-depth prefix-char)))
+
+(defun ellama-get-session-file-extension ()
+  "Return file extension based om the current mode.
+Defaults to md, but supports org.  Depends on \"ellama-major-mode.\""
+  (cond ((provided-mode-derived-p ellama-major-mode 'org-mode) "org")
+        (t "md")))
+
 (defun ellama-new-session (provider prompt &optional ephemeral)
   "Create new ellama session with unique id.
 Provided PROVIDER and PROMPT will be used in new session.
@@ -441,14 +530,15 @@ If EPHEMERAL non nil new session will not be associated with any file."
 			       ellama-session-auto-save)
 		      (file-name-concat
 		       ellama-sessions-directory
-		       (concat id "." ellama-session-file-extension))))
+		       (concat id "." (ellama-get-session-file-extension)))))
 	 (session (make-ellama-session
-		   :id id :provider provider :file file-name))
+		   :id id :provider provider :file file-name :context ellama--new-session-context))
 	 (buffer (if file-name
 		     (progn
 		       (make-directory ellama-sessions-directory t)
 		       (find-file-noselect file-name))
 		   (get-buffer-create id))))
+    (setq ellama--new-session-context nil)
     (setq ellama--current-session-id id)
     (puthash id buffer ellama--active-sessions)
     (with-current-buffer buffer
@@ -490,6 +580,19 @@ If EPHEMERAL non nil new session will not be associated with any file."
 	   (concat "." base-name ".session.el"))))
     session-file-name))
 
+(defun ellama--get-translation-file-name (file-name)
+  "Get ellama translation file name for FILE-NAME."
+  (let* ((base-name (file-name-base file-name))
+	 (ext (file-name-extension file-name))
+	 (dir (file-name-directory file-name))
+	 (translation-file-name
+	  (file-name-concat
+	   dir
+	   (concat base-name ".translation"
+		   (when ext
+		     (concat "." ext))))))
+    translation-file-name))
+
 (defun ellama--save-session ()
   "Save current ellama session."
   (when ellama--current-session
@@ -541,7 +644,15 @@ If EPHEMERAL non nil new session will not be associated with any file."
 	(save-buffer)
 	(goto-char (point-min))))
     (with-current-buffer buffer
-      (setq ellama--current-session (read session-buffer))
+      (let ((session (read session-buffer)))
+	(setq ellama--current-session
+	      (make-ellama-session
+	       :id (ellama-session-id session)
+	       :provider (ellama-session-provider session)
+	       :file (ellama-session-file session)
+	       :prompt (ellama-session-prompt session)
+	       :context ellama--new-session-context)))
+      (setq ellama--new-session-context nil)
       (setq ellama--current-session-id (ellama-session-id ellama--current-session))
       (puthash (ellama-session-id ellama--current-session)
 	       buffer ellama--active-sessions)
@@ -558,10 +669,20 @@ If EPHEMERAL non nil new session will not be associated with any file."
 	      (hash-table-keys ellama--active-sessions)))
 	 (buffer (ellama-get-session-buffer id))
 	 (file (buffer-file-name buffer))
-	 (session-file (ellama--get-session-file-name file)))
+	 (session-file (ellama--get-session-file-name file))
+	 (translation-file (ellama--get-translation-file-name file)))
     (kill-buffer buffer)
     (delete-file file t)
-    (delete-file session-file t)))
+    (delete-file session-file t)
+    (mapc
+     (lambda (buf)
+       (when (and (buffer-file-name buf)
+		  (file-equal-p (buffer-file-name buf)
+				translation-file))
+	 (kill-buffer buf)))
+     (buffer-list))
+    (when (file-exists-p translation-file)
+      (delete-file translation-file t))))
 
 ;;;###autoload
 (defun ellama-session-switch ()
@@ -608,6 +729,134 @@ If EPHEMERAL non nil new session will not be associated with any file."
     (remhash id ellama--active-sessions)
     (puthash new-id buffer ellama--active-sessions)))
 
+;;;###autoload
+(defun ellama-context-add-file ()
+  "Add file to context."
+  (interactive)
+  (if-let* ((id ellama--current-session-id)
+	    (session (with-current-buffer (ellama-get-session-buffer id)
+		       ellama--current-session))
+	    (file-name (read-file-name "Select file: " nil nil t)))
+      (push (cons 'file file-name) (ellama-session-context session))
+    (push (cons 'file file-name) ellama--new-session-context)))
+
+;;;###autoload
+(defun ellama-context-add-buffer (buf)
+  "Add BUF to context."
+  (interactive "bSelect buffer: ")
+  (if-let* ((id ellama--current-session-id)
+	    (session (with-current-buffer (ellama-get-session-buffer id)
+		       ellama--current-session)))
+      (push (cons 'buffer buf) (ellama-session-context session))
+    (push (cons 'buffer buf) ellama--new-session-context)))
+
+;;;###autoload
+(defun ellama-context-add-selection ()
+  "Add file to context."
+  (interactive)
+  (if-let* ((id ellama--current-session-id)
+	    (session (with-current-buffer (ellama-get-session-buffer id)
+		       ellama--current-session))
+	    ((region-active-p))
+	    (content (buffer-substring-no-properties (region-beginning) (region-end))))
+      (push (cons 'text content) (ellama-session-context session))
+    (push (cons 'text content) ellama--new-session-context)))
+
+;;;###autoload
+(defun ellama-context-add-info-node (node)
+  "Add info NODE to context."
+  (interactive (list (Info-copy-current-node-name)))
+  (if-let* ((id ellama--current-session-id)
+	    (session (with-current-buffer (ellama-get-session-buffer id)
+		       ellama--current-session)))
+      (push (cons 'info node) (ellama-session-context session))
+    (push (cons 'info node) ellama--new-session-context)))
+
+(defun ellama--translate-string (s)
+  "Translate string S to `ellama-language' syncronously."
+  (llm-chat
+   (or ellama-translation-provider ellama-provider)
+   (llm-make-simple-chat-prompt
+    (format ellama-translation-template
+	    ellama-language
+	    s
+	    ellama-language))))
+
+(defun ellama--org-format-context-element (elt)
+  "Format context ELT for org mode."
+  (pcase (car elt)
+    ('file
+     (format "[[file:%s][%s]]" (cdr elt) (cdr elt)))
+    ('buffer
+     (format "[[elisp:(display-buffer \"%s\")][%s]]" (cdr elt) (cdr elt)))
+    ('text
+     (cdr elt))
+    ('info
+     (format "[[%s][%s]]"
+	     (replace-regexp-in-string
+	      "(\\(.?*\\)) \\(.*\\)" "info:\\1#\\2" (cdr elt))
+	     (if (and ellama-chat-translation-enabled
+		      (not ellama--current-session))
+		 (ellama--translate-string (cdr elt))
+	       (cdr elt))))
+    (_
+     (user-error "Unsupported context element"))))
+
+(defun ellama--md-format-context-element (elt)
+  "Format context ELT for org mode."
+  (pcase (car elt)
+    ('file
+     (format "[%s](<%s>)"
+	     (cdr elt)
+	     (cdr elt)))
+    ('buffer
+     (format "```emacs-lisp\n(display-buffer \"%s\")\n```\n" (cdr elt)))
+    ('text
+     (cdr elt))
+    ('info
+     (format "```emacs-lisp\n(info \"%s\")\n```\n" (cdr elt)))
+    (_
+     (user-error "Unsupported context element"))))
+
+(defun ellama--extract-context-element (elt)
+  "Extract context ELT content."
+  (pcase (car elt)
+    ('text (cdr elt))
+    ('buffer (with-current-buffer (cdr elt)
+	       (buffer-substring-no-properties (point-min) (point-max))))
+    ('file (with-temp-buffer
+	     (find-file-literally (cdr elt))
+	     (buffer-substring-no-properties (point-min) (point-max))))
+    ('info (with-temp-buffer
+	     (info (cdr elt) (current-buffer))
+	     (buffer-substring-no-properties (point-min) (point-max))))
+    (_
+     (user-error "Unsupported context element"))))
+
+(defun ellama--format-context (session)
+  "Format SESSION context for chat buffer."
+  (if-let* ((context (ellama-session-context session)))
+      (concat (string-join
+	       (cons "Context:"
+		     (if (derived-mode-p 'org-mode)
+			 (mapcar #'ellama--org-format-context-element context)
+		       (mapcar #'ellama--md-format-context-element context)))
+	       "\n")
+	      "\n\n")
+    ""))
+
+(defun ellama--prompt-with-context (prompt)
+  "Add context to PROMPT for sending to llm."
+  (if-let* ((session ellama--current-session)
+	    (context (ellama-session-context session)))
+      (concat (string-join
+	       (cons "Context:"
+		     (mapcar #'ellama--extract-context-element context))
+	       "\n")
+	      "\n\n"
+	      prompt)
+    prompt))
+
 (defun ellama-stream (prompt &rest args)
   "Query ellama for PROMPT.
 ARGS contains keys for fine control.
@@ -632,8 +881,10 @@ failure (with BUFFER current).
 
 :on-done ON-DONE -- ON-DONE a function that's called with the full response text
 when the request completes (with BUFFER current)."
-  (let* ((provider (or (plist-get args :provider) ellama-provider))
-	 (session (plist-get args :session))
+  (let* ((session (plist-get args :session))
+	 (provider (if session
+		       (ellama-session-provider session)
+		     (or (plist-get args :provider) ellama-provider)))
 	 (buffer (or (plist-get args :buffer)
 		     (when (ellama-session-p session)
 		       (ellama-get-session-buffer (ellama-session-id session)))
@@ -645,16 +896,17 @@ when the request completes (with BUFFER current)."
 		    (lambda (msg)
 		      (error "Error calling the LLM: %s" msg))))
 	 (donecb (or (plist-get args :on-done) #'ignore))
+	 (prompt-with-ctx (ellama--prompt-with-context prompt))
 	 (llm-prompt (if session
 			 (if (llm-chat-prompt-p (ellama-session-prompt session))
 			     (progn
 			       (llm-chat-prompt-append-response
 				(ellama-session-prompt session)
-				prompt)
+				prompt-with-ctx)
 			       (ellama-session-prompt session))
 			   (setf (ellama-session-prompt session)
-				 (llm-make-simple-chat-prompt prompt)))
-		       (llm-make-simple-chat-prompt prompt))))
+				 (llm-make-simple-chat-prompt prompt-with-ctx)))
+		       (llm-make-simple-chat-prompt prompt-with-ctx))))
     (with-current-buffer buffer
       (ellama-request-mode +1)
       (let* ((start (make-marker))
@@ -690,6 +942,8 @@ when the request completes (with BUFFER current)."
 	(set-marker-insertion-type start nil)
 	(set-marker-insertion-type end t)
 	(spinner-start ellama-spinner-type)
+	(when session
+	  (setf (ellama-session-context session) nil))
 	(setq ellama--current-request
 	      (llm-chat-streaming provider
 				  llm-prompt
@@ -721,58 +975,128 @@ Will call `ellama-chat-done-callback' on TEXT."
   (when ellama-chat-done-callback
     (funcall ellama-chat-done-callback text)))
 
-;;;###autoload
-(defun ellama-chat (prompt &optional create-session)
-  "Send PROMPT to ellama chat with conversation history.
+(defun ellama--translate-generated-text-on-done (translation-buffer)
+  "Translate generated text into TRANSLATION-BUFFER."
+  (lambda (generated)
+    (ellama-chat-done generated)
+    (display-buffer translation-buffer)
+    (with-current-buffer translation-buffer
+      (save-excursion
+	(goto-char (point-max))
+	(ellama-stream
+	 (format ellama-translation-template
+		 ellama-language
+		 generated
+		 ellama-language)
+	 :provider (or ellama-translation-provider ellama-provider)
+	 :on-done #'ellama-chat-done
+	 :filter (when (derived-mode-p 'org-mode)
+		   #'ellama--translate-markdown-to-org-filter))))))
 
-If CREATE-SESSION set, creates new session even if there is an active session."
-  (interactive "sAsk ellama: ")
-  (let* ((providers (progn
-		      (push '("default model" . ellama-provider)
-			    ellama-providers)
-		      (if (and ellama-ollama-binary
-			       (file-exists-p ellama-ollama-binary))
-			  (push '("ollama model" . (ellama-get-ollama-local-model))
-				ellama-providers)
-			ellama-providers)))
-	 (variants (mapcar #'car providers))
-	 (provider (if current-prefix-arg
-		       (eval (alist-get
-			      (completing-read "Select model: " variants)
-			      providers nil nil #'string=))
-		     ellama-provider))
-	 (session (if (or create-session
-			  current-prefix-arg
-			  (and (not ellama--current-session)
-			       (not ellama--current-session-id)))
-		      (ellama-new-session provider prompt)
-		    (or ellama--current-session
-			(with-current-buffer (ellama-get-session-buffer
-					      ellama--current-session-id)
-			  ellama--current-session))))
-	 (buffer (ellama-get-session-buffer
-		  (ellama-session-id session))))
+(defun ellama--call-llm-with-translated-prompt (buffer session translation-buffer)
+  "Call llm with translated text in BUFFER with SESSION from TRANSLATION-BUFFER."
+  (lambda (result)
+    (ellama-chat-done result)
+    (save-excursion
+      (goto-char (point-max))
+      (delete-char -2)
+      (delete-char (- (length result))))
     (display-buffer buffer)
     (with-current-buffer buffer
       (save-excursion
 	(goto-char (point-max))
-	(insert ellama-nick-prefix " " ellama-user-nick ":\n" prompt "\n\n"
-		ellama-nick-prefix " " ellama-assistant-nick ":\n")
-	(ellama-stream prompt
+	(insert (ellama-get-nick-prefix-for-mode) " " ellama-user-nick ":\n"
+		(ellama--format-context session) result "\n\n"
+		(ellama-get-nick-prefix-for-mode) " " ellama-assistant-nick ":\n")
+	(ellama-stream result
 		       :session session
-		       :on-done #'ellama-chat-done
+		       :on-done (ellama--translate-generated-text-on-done translation-buffer)
 		       :filter (when (derived-mode-p 'org-mode)
 				 #'ellama--translate-markdown-to-org-filter))))))
 
+(defun ellama--translate-interaction (prompt translation-buffer buffer session)
+  "Translate chat PROMPT in TRANSLATION-BUFFER for BUFFER with SESSION."
+  (display-buffer translation-buffer)
+  (with-current-buffer translation-buffer
+    (save-excursion
+      (goto-char (point-max))
+      (insert (ellama-get-nick-prefix-for-mode) " " ellama-user-nick ":\n"
+	      (ellama--format-context session) prompt "\n\n"
+	      (ellama-get-nick-prefix-for-mode) " " ellama-assistant-nick ":\n")
+      (ellama-stream
+       (format ellama-translation-template
+	       "english"
+	       prompt
+	       "english")
+       :provider (or ellama-translation-provider ellama-provider)
+       :filter (when (derived-mode-p 'org-mode)
+		 #'ellama--translate-markdown-to-org-filter)
+       :on-done
+       (ellama--call-llm-with-translated-prompt buffer session translation-buffer)))))
+
+;;;###autoload
+(defun ellama-chat (prompt &optional create-session &rest args)
+  "Send PROMPT to ellama chat with conversation history.
+
+If CREATE-SESSION set, creates new session even if there is an active session.
+ARGS contains keys for fine control.
+
+:provider PROVIDER -- PROVIDER is an llm provider for generation."
+  (interactive "sAsk ellama: ")
+  (let* ((providers (append
+                     `(("default model" . ellama-provider)
+		               ,(if (and ellama-ollama-binary (file-exists-p ellama-ollama-binary))
+			                '("ollama model" . (ellama-get-ollama-local-model))))
+                     ellama-providers))
+	     (variants (mapcar #'car providers))
+	     (provider (if current-prefix-arg
+		               (eval (alist-get
+			                  (completing-read "Select model: " variants)
+			                  providers nil nil #'string=))
+		             (or (plist-get args :provider)
+			             ellama-provider)))
+	     (session (if (or create-session
+			              current-prefix-arg
+			              (and (not ellama--current-session)
+			                   (not ellama--current-session-id)))
+		              (ellama-new-session provider prompt)
+		            (or ellama--current-session
+			            (with-current-buffer (ellama-get-session-buffer
+					                          ellama--current-session-id)
+			              ellama--current-session))))
+	     (buffer (ellama-get-session-buffer
+		          (ellama-session-id session)))
+	     (file-name (ellama-session-file session))
+	     (translation-buffer (when ellama-chat-translation-enabled
+			                   (if file-name
+				                   (progn
+				                     (find-file-noselect
+				                      (ellama--get-translation-file-name file-name)))
+				                 (get-buffer-create (ellama-session-id session))))))
+    (if ellama-chat-translation-enabled
+	(ellama--translate-interaction prompt translation-buffer buffer session)
+      (display-buffer buffer)
+      (with-current-buffer buffer
+	(save-excursion
+	  (goto-char (point-max))
+	  (insert (ellama-get-nick-prefix-for-mode) " " ellama-user-nick ":\n"
+		  (ellama--format-context session) prompt "\n\n"
+		  (ellama-get-nick-prefix-for-mode) " " ellama-assistant-nick ":\n")
+	  (ellama-stream prompt
+			 :session session
+			 :on-done #'ellama-chat-done
+			 :filter (when (derived-mode-p 'org-mode)
+				   #'ellama--translate-markdown-to-org-filter)))))))
+
 ;;;###autoload
 (defun ellama-ask-about ()
   "Ask ellama about selected region or current buffer."
   (interactive)
-  (let ((input (read-string "Ask ellama about this text: "))
-	(text (if (region-active-p)
-		  (buffer-substring-no-properties (region-beginning) (region-end))
-		(buffer-substring-no-properties (point-min) (point-max)))))
-    (ellama-chat (format ellama-ask-about-prompt-template text input))))
+  (let ((input (read-string "Ask ellama about this text: ")))
+    (if (region-active-p)
+	(ellama-context-add-selection)
+      (ellama-context-add-buffer (buffer-name (current-buffer))))
+    (ellama-chat input)))
 
 ;;;###autoload
 (defun ellama-ask-selection ()
@@ -803,9 +1127,15 @@ If CREATE-SESSION set, creates new session even if there is an active session."
   (let ((text (thing-at-point 'line)))
     (ellama-chat text)))
 
-(defun ellama-instant (prompt)
-  "Prompt ellama for PROMPT to reply instantly."
-  (let* ((buffer-name (ellama-generate-name ellama-provider real-this-command prompt))
+(defun ellama-instant (prompt &rest args)
+  "Prompt ellama for PROMPT to reply instantly.
+
+ARGS contains keys for fine control.
+
+:provider PROVIDER -- PROVIDER is an llm provider for generation."
+  (let* ((provider (or (plist-get args :provider)
+		       ellama-provider))
+	 (buffer-name (ellama-generate-name provider real-this-command prompt))
 	 (buffer (get-buffer-create (if (get-buffer buffer-name)
 					(make-temp-name (concat buffer-name " "))
 				      buffer-name)))
@@ -817,20 +1147,30 @@ If CREATE-SESSION set, creates new session even if there is an active session."
     (display-buffer buffer)
     (ellama-stream prompt
 		   :buffer buffer
-		   :filter filter)))
+		   :filter filter
+		   :provider provider)))
 
 ;;;###autoload
 (defun ellama-translate ()
   "Ask ellama to translate selected region or word at point."
   (interactive)
-  (if (region-active-p)
-      (ellama-instant
-       (format ellama-translate-region-prompt-template
-	       ellama-language
-	       (buffer-substring-no-properties (region-beginning) (region-end))))
+  (let ((text (if (region-active-p)
+		  (buffer-substring-no-properties (region-beginning) (region-end))
+		(thing-at-point 'word))))
     (ellama-instant
-     (format ellama-translate-word-prompt-template
-	     (thing-at-point 'word) ellama-language))))
+     (format ellama-translation-template
+	     ellama-language text ellama-language)
+     :provider ellama-translation-provider)))
+
+;;;###autoload
+(defun ellama-translate-buffer ()
+  "Ask ellama to translate current buffer."
+  (interactive)
+  (let ((text (buffer-substring-no-properties (point-min) (point-max))))
+    (ellama-instant
+     (format ellama-translation-template
+	     ellama-language text ellama-language)
+     :provider ellama-translation-provider)))
 
 ;;;###autoload
 (defun ellama-define-word ()
@@ -1032,23 +1372,39 @@ buffer."
 		(llm-ollama-host ellama-provider)))
 	(port (when (llm-ollama-p ellama-provider)
 		(llm-ollama-port ellama-provider))))
-    (make-llm-ollama
-     :chat-model model-name :embedding-model model-name :host host :port port)))
+    (if host
+	(make-llm-ollama
+	 :chat-model model-name :embedding-model model-name :host host :port port)
+      (make-llm-ollama
+       :chat-model model-name :embedding-model model-name))))
 
 ;;;###autoload
 (defun ellama-provider-select ()
   "Select ellama provider."
   (interactive)
-  (let* ((providers (if (and ellama-ollama-binary
-			     (file-exists-p ellama-ollama-binary))
-			(push '("ollama model" . (ellama-get-ollama-local-model))
-			      ellama-providers)
-		      ellama-providers))
-	 (variants (mapcar #'car providers)))
+  (let* ((providers (append
+                     `(("default model" . ellama-provider)
+		               ,(if (and ellama-ollama-binary (file-exists-p ellama-ollama-binary))
+			                '("ollama model" . (ellama-get-ollama-local-model))))
+                     ellama-providers))
+	     (variants (mapcar #'car providers)))
     (setq ellama-provider
 	  (eval (alist-get
 		 (completing-read "Select model: " variants)
-		 providers nil nil #'string=)))))
+		 providers nil nil #'string=)))
+    (setq ellama--current-session-id nil)))
+
+;;;###autoload
+(defun ellama-chat-translation-enable ()
+  "Enable chat translation."
+  (interactive)
+  (setq ellama-chat-translation-enabled t))
+
+;;;###autoload
+(defun ellama-chat-translation-disable ()
+  "Enable chat translation."
+  (interactive)
+  (setq ellama-chat-translation-enabled nil))
 
 (provide 'ellama)
 ;;; ellama.el ends here.
